from __future__ import division

import random
import math

import pygame

import data
import vector
import util
import Pathfinder

class Object(object):
    selectable = False
    def __init__(self, pos):
        self.pos = pos
    # dt is delta time, or change in time since the last loop.
    def update(self, dt, world):
        pass
    def draw(self, surf, level, selected):
        if level != self.level: return
        # gets the image name and caches it using util.cdict
        image = data.image[self.image_name]
        surf.blit(image, self.pos-vector.v(image.get_size())/2)

class Building(Object):
    pass

class Hub(Building):
    selectable = True
    image_name = "hub.png"
    level = 2
    health = 100
    def __init__(self, *args):
        Building.__init__(self, *args)
        self.health_counter = self.health
    def update(self, dt, world):
        Building.update(self, dt, world)
        if self.health_counter <= 0:
            world.objects.remove(self)
            return
        self.health_counter += dt*2
        self.health_counter = min(self.health_counter, self.health)
        # tells itself how much string it has to use
        self.string_cache = world.string
    # click handling, yay!
    def on_action(self, pos, world):
        if (pos - self.pos).mag() > world.string:
            world.objects.append(Error('Not enough string'))
            return
        # you can't make segments too short or hubs too close together
        if (pos - self.pos).mag() < 20:
            world.objects.append(Error('String is too short'))
            return
        # this prevents you from building too close to the spawn.
        if min((x.pos-self.pos).mag() for x in world.objects if isinstance(x, Spawn)) < 30:
            world.objects.append(Error('Too close to spawn'))
            return
        try:
            # and this is to prevent building next to an enemy.
            # the except statement catches times when there are no enemies
            if min((x.pos-self.pos).mag() for x in world.objects if isinstance(x, Enemy)) < 10:
                world.objects.append(Error('Too close to enemy'))
                return
        except ValueError: pass
        # can't build too close to any other hub
        if min((x.pos-pos).mag() for x in world.objects if isinstance(x, Hub)) < 30:
            world.objects.append(Error('Too close to another hub'))
            return
        # checks for segment collision and prevents it
        for x in world.objects:
            if not isinstance(x, Segment): continue
            if util.line_intersection([x.pos for x in x.hubs], (self.pos, pos)) is not None:
                world.objects.append(Error('String crosses string'))
                return
        # if you managed to get through all those checks for spawn/enemy/hub proximity,
        # congrats, you have a new hub
        hub = Hub(pos)
        # subtracts your string from your treasury to build the hub
        world.string -= (pos - self.pos).mag()
        # tell the world we have a new segment spanning between the old and new hubs
        world.objects.append(Segment((self, hub)))
        # add the hub to the world
        world.objects.append(hub)
        # new hubs come from the new hub by default
        world.selected = hub
    def draw(self, surf, level, selected):
        if level == self.level:
            pygame.draw.circle(surf, (0, 0, 0), self.pos.int(), 10)
            #for i in xrange(2, 20): # forrest realizes there is no pygame.draw.pie 4 hours before pyweek ends
            #    pygame.draw.arc(surf, (255, 0, 0), (self.pos-(i/2,i/2), (i, i)), 0, (self.health-self.health_counter)/self.health*2*math.pi, 1)
            for i in xrange(0, int((1-self.health_counter/self.health)*200)): # forrest realizes above is horrible
                i = (i/200.)-.25
                pygame.draw.aaline(surf, (255, 0, 0), self.pos, self.pos+(10*math.cos(i*math.pi*2), 10*math.sin(i*math.pi*2)))
        if selected and level == 4 and self.string_cache >= 20:
            # the circle of "i can build here"-ness
            pygame.draw.circle(surf, (255, 0, 0), self.pos.int(), int(self.string_cache), 3)

class Spool(Hub):
    level = 2
    def draw(self, surf, level, selected):
        Hub.draw(self, surf, level, selected)
        # just wondering... why would you ever wanna make the level 0, if it's at 2?
        if level == 0 or level == 2:
            image = data.image['spool.png']
            surf.blit(image, self.pos-vector.v(image.get_size())/2+(0, 10*(1-level)))

class Segment(Building):
    level = 1
    range = 40
    damage = 10
    recharge = 1 # seconds between attk
    def __init__(self, hubs):
        # is this line truly necessary?
        assert len(hubs) == 2
        # will this be a tuple of which hubs it has?
        self.hubs = hubs
        self.recharge_counter = 0
    # what does this @ sign do?
    @property
    def pos(self):
        # so it's position is defined by the (x, y) coordinates of its hubs, each divided by two.
        # why?
        return self.hubs[0].pos/2 + self.hubs[1].pos/2
    def update(self, dt, world):
        if not all(x in world.objects for x in self.hubs):
            world.objects.remove(self)
            return
        # imma chargin my lazor...
        self.recharge_counter -= dt
        if self.recharge_counter < 0:
            potential = []
            for object in world.objects:
                # i'm guessing isinstance() is built in.
                if not isinstance(object, Enemy): continue
                # uses that super crazy expression to find the closest point, but what if there's two?
                attk_pt = util.closest_point_on_line([x.pos for x in self.hubs], object.pos)
                # i still dont get what .mag() is.
                dist = (object.pos-attk_pt).mag()
                # code no longer needed, i assume
                '''
                pygame.draw.circle(pygame.display.get_surface(), (0, 255, 0), attk_pt, 5)
                pygame.display.update()
                '''
                # if the monster's too far to attack, forget it
                if dist > self.range: continue
                # otherwise, we add it to our list of victims
                potential.append(object)
            # sort out which victim would be best to kill
            potential.sort(key=lambda x:x.health/x.damage)
            # apparently, if potential is empty, this block isn't executed
            # but the interpreter tells me a blank list is neither True, False, nor None
            # wtf?
            if potential:
                # takes the first enemy off the sorted list to kill
                enemy = potential[0]
                # why does it find the closest attack point again?
                dir = (self.hubs[1].pos-self.hubs[0].pos).norm()
                attk_pt = util.closest_point_on_line((self.hubs[0].pos+dir*10, self.hubs[1].pos-dir*10), enemy.pos)
                # die enemy!
                enemy.health_counter -= self.damage
                # bzzzt!
                world.objects.append(Attack(attk_pt, enemy.pos, (255, 255, 0)))
                # whoops, you can't attack for a while
                self.recharge_counter += self.recharge
        # lasers ready to fire cap'n!
        if self.recharge_counter < 0: self.recharge_counter = 0
    def draw(self, surf, level, selected):
        if level != self.level: return
        # hehe, lines are white
        pygame.draw.aaline(surf, (255,255-255*self.recharge_counter/self.recharge,255-255*self.recharge_counter/self.recharge), self.hubs[0].pos, self.hubs[1].pos)

class Spawn(Object):
    image_name = "spawn.png"
    level = 0
    def __init__(self, *args):
        Object.__init__(self, *args)
        # first enemy pops up in 5 seconds
        self.countdown = 5
        self.counter = 0
    def update(self, dt, world):
        self.countdown -= dt
        while self.countdown < 0:
            self.counter += 1
            enemy = Enemy(self.pos)
            enemy.health += self.counter
            enemy.damage += 0.1 * self.counter
            # -bloop- A wild Enemy hath appeared!
            world.objects.append(enemy)
            # another enemy will come up anywhere from 3-6 seconds later
            self.countdown += random.uniform(3, 6)

class Enemy(Object):
    # i like my enemy3.png better than the original
    image_name = 'enemy3.png'
    level = 1
    health = 50
    damage = 5
    range = 20
    recharge = 1
    def __init__(self, *args):
        Object.__init__(self, *args)
        self.health_counter = self.health
        self.recharge_counter = 0
    def update(self, dt, world):
        # die die die
        if self.health_counter <= 0:
            world.objects.remove(self)
            # hahahahaha i get string!
            world.string += 40
            # end function! you dead!
            return
        # charging his lazor
        self.recharge_counter -= dt
        # find nearest hub
        dest = min((x for x in world.objects if isinstance(x, Hub)), key=lambda x: (x.pos-self.pos).mag())
        path = enemy_pathfinder.run(self.pos, [x for x in world.objects if isinstance(x, Spawn)][0].pos, world)
        
        # go towards hub?
        dir = dest.pos-self.pos
        # still, wtf.
        dist = dir.mag()
        if path:
            dest2 = path[1]
            dir2 = dest2.pos-self.pos
            dist2 = dir.mag()
            dir2 = dir2 / dist2
            speed2 = min(20, self.range/dt)
            self.pos2 = self.pos + dir2*speed2*dt
            self.pos = self.pos + vector.v(-dir2[1], dir2[0])*random.uniform(-3, 3)*dt
        elif dist > self.range:
            dir = dir / dist
            speed = min(20, self.range/dt)
            self.pos = self.pos + dir*speed*dt
            self.pos = self.pos + vector.v(-dir[1], dir[0])*random.uniform(-3, 3)*dt
        elif self.recharge_counter < 0:
            potential = []
            for object in world.objects:
                if not isinstance(object, Hub): continue
                attk_pt = self.pos
                dist = (object.pos-attk_pt).mag()
                #pygame.draw.circle(pygame.display.get_surface(), (0, 255, 0), attk_pt, 5)
                #pygame.display.update()
                if dist > self.range: continue
                potential.append(object)
            potential.sort(key=lambda x:x.health*(x.pos-self.pos).mag())
            if potential:
                enemy = potential[0]
                attk_pt = self.pos
                enemy.health_counter -= self.damage
                world.objects.append(Attack(attk_pt, enemy.pos, (255, 0, 0)))
                self.recharge_counter += self.recharge
        if self.recharge_counter < 0: self.recharge_counter = 0

class EnemyPathfinder(Pathfinder.LinearPathfinder):
    def children(self, node, world):
        res = []
        for n in world.objects:
            if not isinstance(n, Spawn): continue
            good = True
            for oth in world.objects:
                if not isinstance(oth, Segment): continue
                if util.line_intersection((node, n.pos), [hu.pos for hu in oth.hubs]):
                    good = False
            if good:
                res.append(n.pos)
        return res
enemy_pathfinder = EnemyPathfinder()

class Effect(Object):
    level = 3
    def __init__(self, *args):
        Object.__init__(self, *args)
        self.timeleft = self.time
    def update(self, dt, world):
        self.timeleft -= dt
        if not self.timeleft > 0:
            world.objects.remove(self)

class Attack(Effect):
    time = .5
    def __init__(self, start, end, color):
        try:
            data.sound['light.wav'].play()
        except pygame.error:
            pass
        Effect.__init__(self, start/2+end/2)
        self.start = start
        self.end = end
        self.color = color
    def draw(self, surf, level, selected):
        if level != self.level: return
        dir = (self.end-self.start).norm()
        perp = vector.v(-dir[1], dir[0])
        x = 0
        y = 0
        while x < (self.start-self.end).mag():
           ox = x
           oy = y
           x += random.uniform(3, 5)
           y = random.gauss(0, 3)
           pygame.draw.aaline(surf, self.color, self.start+ox*dir+oy*perp, self.start+x*dir+y*perp)

class Error(Effect):
    time = 3
    def __init__(self, message):
        Effect.__init__(self, None)
        try:
            data.sound['error.wav'].play()
        except:
            pass
        self.message = message
    def update(self, dt, world):
        for o in world.objects:
           if isinstance(o, Error) and o is not self:
               world.objects.remove(o)
        Effect.update(self, dt, world)
        self.pos = world.size*(.5, .9)
    def draw(self, surf, level, selected):
        if level != self.level: return
        image = data.font[20].render(self.message, True, (255, 255, 0))
        surf.blit(image, self.pos-vector.v(image.get_size())/2)

class World(object):
    def __init__(self, size):
        self.size = size
        self.objects = list()
        self.objects.append(Spool((20, 20)+(180,180)*vector.random(2)))
        self.objects.append(Spawn((312,312)+(180,180)*vector.random(2)))
        self.selected = self.objects[0]
        self.string = 300
    def on_click(self, pos, button):
        if button == 1: # select
            if not self.objects: return
            self.selected = min((x for x in self.objects if x.selectable), key=lambda x:(x.pos-pos).mag())
        elif button == 3:
            if not self.selected: return
            self.selected.on_action(pos, self)
    def update(self, dt):
        try:
          for object in list(self.objects): # copied so objects can delete themselves
            object.update(dt, self)
        except:
          return True
        if len([x for x in self.objects if isinstance(x, Spool)]) == 0:
          return True
    def draw(self, surf):
        """
        levels:
            ground
            0 - things flat on ground
            1 - things above ground
            2 - highest ground things
            3 - effects
            4 - selection
        """
        surf.blit(data.image['bg.png'], (0, 0))
        self.objects.sort(key=lambda x: x.pos[1])
        for level in xrange(5):
            for object in self.objects:
                object.draw(surf, level, object is self.selected)
